#if defined(_WIN32) || defined(NATIVE)
#include <cmath>
#include <fstream>
#include <iostream>
#include <stdexcept> // std::runtime_error
#include <stdio.h>
#include <string>
#include <vector>
#elif defined(TEENSY)
#include <Arduino.h>
#endif

#define M_PI 3.14159265359

#include "Vehicle.h"
#include "sim.h"

Vehicle State;
Vehicle PrevState;

#if defined(TEENSY)

int BUILTIN_LED = 13;
unsigned long last, initTime;

#include "teensy.h"
#include <HX711.h>
#include <PWMServo.h>

HX711 lc0;
HX711 lc1;
HX711 lc2;
HX711 lc3;

void read_lc0();
void read_lc1();
void read_lc2();
void read_lc3();

PWMServo servo1;
PWMServo servo2;

const int pin_servo1 = 33;
const int pin_servo2 = 29;

int servoPos = 0; // variable to store the servo position;

void setup() {
  Serial.begin(9600);

  delay(5000);
  Serial.println("Simulation Countdown:");
  for (int i = 0; i < 10; i++) {
    Serial.println(10 - i);
    delay(1000);
  }

  init_loadCells(lc0, lc1, lc2, lc3);
  init_Vehicle(State);
  State.lc0 = lc0.get_value();
  State.lc1 = lc1.get_value();
  State.lc2 = lc2.get_value();
  State.lc3 = lc3.get_value();

  servo1.attach(pin_servo1);
  servo2.attach(pin_servo2);
  testGimbal(servo1, servo2);
  Serial.println("Servos Tested");

  delay(3000);
  Serial.println("Simulated Vehicle Initalized");

  // Determine when to burn
  burnStartTimeCalc(State);
  Serial.println("Starting Height Calculated");

  initFile();

  initTime = micros();
}

void loop() {
  last = micros();

  vehicleDynamics(State, PrevState);
  thrustInfo(State);
  pidController(State, PrevState);
  TVC(State, PrevState);
  processTVC(State, servo1, servo2);

  State.stepDuration = micros() - last;
  write2CSV(State);

  // Set previous values for next timestep
  PrevState = State;
  State.time += State.stepSize;

  if ((State.z < 0.0) && (State.thrustFiring == 2)) {
    printSimResults(State);
    Serial.println("Run duration:" + String((micros() - initTime) / 1000000.0) +
                   " seconds");

    closeFile();
    delay(3000);

    Serial.println("SUCCESS");
    Serial.println("Exiting Sim");
    teensyAbort();
  }

  delay(State.stepSize - ((micros() - last) / 1000.0));
}

// ISRs to print data to serial monitor - These unfortunately can't move
void read_lc0() { State.lc0 = lc0.get_value(); }
void read_lc1() { State.lc1 = lc1.get_value(); }
void read_lc2() { State.lc2 = lc2.get_value(); }
void read_lc3() { State.lc3 = lc3.get_value(); }

#endif

#if defined(_WIN32) || defined(NATIVE)

#include "native.h"
outVector stateVector;

int main() {

  init_Vehicle(State);

  // Determine when to burn
  burnStartTimeCalc(State);

  do {
    vehicleDynamics(State, PrevState);
    thrustInfo(State);
    pidController(State, PrevState);
    TVC(State, PrevState);
    processTVC(State);
    state2vec(State, PrevState, stateVector);

    // Set "prev" values for next timestep
    PrevState = State;
    State.time += State.stepSize;

    if ((State.z < 0.0) && (State.thrustFiring == 2)) {
      write2CSV(stateVector, State);
      printSimResults(State);
      init_Vehicle(State);
    }
  } while ((State.z > 0.0) || (State.thrustFiring != 2));

  return 0;
}
#endif
